#options list
#options title="Grammar for C- scripting language"

// Semantic actions for language pieces.

procedure Script_Semantics for 

// These semantic names must come first, because their numbers are used to
// index into the error text strings.
	   
	Missing_Comment_End, Missing_Right_Paren,
	Missing_Closing_Quote, Statement_Expected, Invalid_String_Escape,
	Value_Expected, Variable_Name_Expected, Missing_Right_Brace, Invalid_Function,
	
	Push, Pop, Pop_Clean, 
	The_End, 
	Switch_Test, Switch_Off, Test_While, Do_Test,
	Less_Than, Greater_Than, Not_Equal, Equal,
	Init_String, String_Seg1, String_Seg2, String_Seg3, String_Octal,
	Push_Integer,
	Push_Variable, Assign_Variable, Var_Name, Local_Variable, Function_Name,
	Add, Subtract, Multiply, Divide, And, Or,
	Function;

//
// The following grammar is used to implement an interpreter for C-, a scripting
// language used for logging in and out if using a serial connection such as to the
// POP, SMTP and PH servers from Eudora or NewsWatcher.
//

goal <script> 		is [<white_space> <Script_Command>]... <white_space>
						FENCE {@The_End or @Statement_Expected};

<Script_Command>	is	<block_statement>	
					or	<statement> [";"] 
					or	";" ;
					
<Need_Command>		is	<Script_Command>
					or	FENCE @Statement_Expected;
					
<block_statement>	is	<block>
					or	<if>
					or	<while>;
					
<statement>			is	<local_variables>
					or	<assignment>
					or	<function>			@Pop_Clean;
					
<assignment>		is	<variable> <white_space> "=" FENCE
						@Push <expression> @Pop @Assign_Variable;

<comment>			is	"#" <rest_of_line> <line_end>
					or	"/*" [<no_star_text> or <just_asterisk>]...
							{"*/" or FENCE @Missing_Comment_End} ;
			
<just_asterisk>		is	"*/" FAIL
					or	"*";

<if>				is	"if" <sep> FENCE @Push
							<expression> @Do_Test <Need_Command> 
							@Switch_Test <white_space>
						["else" <white_space> "if" <sep> FENCE @Push
							<expression> @Do_Test <Need_Command>
							@Switch_Off @Pop_Clean <white_space>
						]...
						["else" <sep> <white_space> FENCE <Need_Command>]
						@Pop_Clean;

<while>				is	"while" <sep> FENCE @Push
							<expression> @Do_Test <Need_Command>
							@Test_While @Pop_Clean;
				
<block>				is	"{" FENCE @Push
							[ <white_space> <Script_Command> ]... <white_space>
							FENCE { "}" or @Missing_Right_Brace} @Pop_Clean;
				
<local_variables>	is	"variable" <sep> <white_space> 
							<local_variable> <white_space>
							[ "," FENCE <white_space> <local_variable> 
								<white_space> ]...;
					
<function>			is	<function_name> @Push @Function_Name 
						<parameters> FENCE {@Function or @Invalid_Function};
			
<function_name>		is	<letter> [<id_part>];
<variable>			is	<letter> [<id_part>] <no_parameters> @Var_Name;
<local_variable>	is	<letter> [<id_part>] @Local_Variable
					or	FENCE @Variable_Name_Expected;
				
<parameters>		is	<white_space> "(" <white_space> FENCE 
						{	")"
						 or
						 	<expression> [ "," FENCE <expression>]...
							FENCE { ")" or @Missing_Right_Paren}
						};
						
<no_parameters>		is	<white_space> "(" FAIL
					or	SUCCEED;

<value>				is	<integer>
					or	<string>
					or	<variable> @Push_Variable
					or	<function> @Pop
					or	@Value_Expected;
			
<expression>		is	<logical_formula> 
						[  "&" FENCE <logical_formula> @And
						or "|" FENCE <logical_formula> @Or ]...;
				
<logical_formula>	is	<arithmetic_expression>
						[  "<" FENCE <arithmetic_expression> @Less_Than
						or ">" FENCE <arithmetic_expression> @Greater_Than
						or "!=" FENCE <arithmetic_expression> @Not_Equal
						or "=" ["="] FENCE <arithmetic_expression> @Equal
						]...;
					
<arithmetic_expression>
					is	<term>
						[  "+" FENCE <term> @Add
						or "-" FENCE <term> @Subtract ]...;
		
<term>				is	<factor>
						[  "*" FENCE <factor> @Multiply
						or "/" FENCE <factor> @Divide ]...;
		
<factor>			is	<white_space>
						{  "(" FENCE <expression> FENCE 
							{ ")" or @Missing_Right_Paren}
						or <value> }
						<white_space>;

<string>			is	"""" FENCE @Init_String [<String_Seg>]... FENCE 
							{ """" or @Missing_Closing_Quote} ;

<String_Seg>		is	<Nonq> @String_Seg1
					or	"""""" @String_Seg2
					or	"\\" 
						{	<Escape_Char> @String_Seg3
						or	<octal> @String_Octal
						or	FENCE @Invalid_String_Escape};
				
<white_space>		is	[<comment> or <ws>]...;
<integer>			is	["+" or "-"] <digits> @Push_Integer;

// <sep> must succeed after "if", "elseif", "else", "while" and "variable" 
// to differentiate them
// from variables that start with those characters.
<sep>				is	<id_part> FAIL
					or	SUCCEED;

terminal <ws> is 1 or more of " \t" '0D0A00';
terminal <rest_of_line> is 0 or more excluding '0D0A';
terminal <line_end> is 0 or more of '0D0A';
terminal <no_star_text> is 1 or more excluding "*";
terminal <digits> is 1 to 15 of "0123456789";
terminal <Nonq> is 1 or more excluding """\\";
terminal <Escape_Char> is 1 of "?'""\\NRTB";
terminal <letter> is 1 of "ABCDEFGHIJKLMNOPQRSTUVWXYZ";
terminal <id_part> is 1 or more of "ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_";
terminal <octal> is 1 to 3 of "01234567";